@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2020, the neonavigation authors
+ * Copyright (c) 2014-2023, the neonavigation authors
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,10 +44,13 @@
 #include <costmap_cspace_msgs/CSpace3D.h>
 #include <costmap_cspace_msgs/CSpace3DUpdate.h>
 #include <diagnostic_updater/diagnostic_updater.h>
+#include <dynamic_reconfigure/server.h>
 #include <geometry_msgs/PoseArray.h>
 #include <nav_msgs/GetPlan.h>
 #include <nav_msgs/OccupancyGrid.h>
 #include <nav_msgs/Path.h>
+#include <neonavigation_metrics_msgs/Metrics.h>
+#include <neonavigation_metrics_msgs/helper.h>
 #include <planner_cspace_msgs/PlannerStatus.h>
 #include <sensor_msgs/PointCloud.h>
 #include <std_srvs/Empty.h>
@@ -65,6 +68,7 @@
 
 #include <neonavigation_common/compatibility.h>
 
+#include <planner_cspace/Planner3DConfig.h>
 #include <planner_cspace/bbf.h>
 #include <planner_cspace/grid_astar.h>
 #include <planner_cspace/jump_detector.h>
@@ -103,6 +107,7 @@ protected:
   ros::Publisher pub_start_;
   ros::Publisher pub_end_;
   ros::Publisher pub_status_;
+  ros::Publisher pub_metrics_;
   ros::ServiceServer srs_forget_;
   ros::ServiceServer srs_make_plan_;
 
@@ -111,6 +116,7 @@ protected:
   planner_cspace_msgs::MoveWithToleranceGoalConstPtr goal_tolerant_;
   tf2_ros::Buffer tfbuf_;
   tf2_ros::TransformListener tfl_;
+  dynamic_reconfigure::Server<Planner3DConfig> parameter_server_;
 
   Astar as_;
   Astar::Gridmap<char, 0x40> cm_;
@@ -150,7 +156,7 @@ protected:
   bool has_start_;
   bool has_hysteresis_map_;
   std::vector<Astar::Vec> hyst_updated_cells_;
-  bool goal_updated_;
+  bool cost_estim_cache_created_;
   bool remember_updates_;
   bool fast_map_update_;
   std::vector<Astar::Vec> search_list_;
@@ -164,6 +170,7 @@ protected:
   float remember_miss_odds_;
   bool use_path_with_velocity_;
   bool retain_last_error_status_;
+  int num_cost_estim_task_;
 
   JumpDetector jump_;
   std::string robot_frame_;
@@ -186,6 +193,7 @@ protected:
   int goal_tolerance_ang_;
 
   planner_cspace_msgs::PlannerStatus status_;
+  neonavigation_metrics_msgs::Metrics metrics_;
 
   bool find_best_;
   float sw_wait_;
@@ -199,6 +207,7 @@ protected:
   bool escaping_;
 
   int cnt_stuck_;
+  bool is_start_occupied_;
 
   diagnostic_updater::Updater diag_updater_;
   ros::Duration costmap_watchdog_;
@@ -405,12 +414,13 @@ protected:
       escaping_ = false;
       has_goal_ = true;
       cnt_stuck_ = 0;
-      if (!updateGoal())
+      if (!createCostEstimCache())
       {
         has_goal_ = false;
         return false;
       }
       status_.status = planner_cspace_msgs::PlannerStatus::DOING;
+      status_.header.stamp = ros::Time::now();
       pub_status_.publish(status_);
       diag_updater_.update();
     }
@@ -480,7 +490,7 @@ protected:
     ROS_DEBUG("    (%d,%d,%d)", s[0], s[1], s[2]);
     return true;
   }
-  bool updateGoal(const bool goal_changed = true)
+  bool createCostEstimCache(const bool goal_changed = true)
   {
     if (!has_goal_)
       return true;
@@ -492,6 +502,9 @@ protected:
       return true;
     }
 
+    cost_estim_cache_created_ = false;
+    is_start_occupied_ = false;
+
     Astar::Vec s, e;
     grid_metric_converter::metric2Grid(
         map_info_, s[0], s[1], s[2],
@@ -500,14 +513,16 @@ protected:
     s.cycleUnsigned(map_info_.angle);
     grid_metric_converter::metric2Grid(
         map_info_, e[0], e[1], e[2],
-        goal_.pose.position.x, goal_.pose.position.y,
-        tf2::getYaw(goal_.pose.orientation));
+        goal_raw_.pose.position.x, goal_raw_.pose.position.y,
+        tf2::getYaw(goal_raw_.pose.orientation));
     e.cycleUnsigned(map_info_.angle);
     if (goal_changed)
     {
-      ROS_INFO(
-          "New goal received (%d, %d, %d)",
-          e[0], e[1], e[2]);
+      ROS_INFO("New goal received. Metric: (%.3f, %.3f, %.3f), Grid: (%d, %d, %d)",
+               goal_raw_.pose.position.x, goal_raw_.pose.position.y, tf2::getYaw(goal_raw_.pose.orientation),
+               e[0], e[1], e[2]);
+      clearHysteresis();
+      has_hysteresis_map_ = false;
     }
     const DiscretePoseStatus start_pose_status = relocateDiscretePoseIfNeeded(s, tolerance_range_, tolerance_angle_);
     const DiscretePoseStatus goal_pose_status = relocateDiscretePoseIfNeeded(e, tolerance_range_, tolerance_angle_);
@@ -518,6 +533,8 @@ protected:
         return false;
       case DiscretePoseStatus::IN_ROCK:
         ROS_WARN("Oops! You are in Rock!");
+        ++cnt_stuck_;
+        is_start_occupied_ = true;
         return true;
       default:
         break;
@@ -532,31 +549,45 @@ protected:
         ++cnt_stuck_;
         return true;
       case DiscretePoseStatus::RELOCATED:
-        ROS_INFO("Goal moved (%d, %d, %d)", e[0], e[1], e[2]);
         float x, y, yaw;
         grid_metric_converter::grid2Metric(map_info_, e[0], e[1], e[2], x, y, yaw);
         goal_.pose.orientation = tf2::toMsg(tf2::Quaternion(tf2::Vector3(0.0, 0.0, 1.0), yaw));
         goal_.pose.position.x = x;
         goal_.pose.position.y = y;
+        ROS_INFO("Goal moved. Metric: (%.3f, %.3f, %.3f), Grid: (%d, %d, %d)", x, y, yaw, e[0], e[1], e[2]);
         break;
       default:
+        Astar::Vec e_prev;
+        grid_metric_converter::metric2Grid(
+            map_info_, e_prev[0], e_prev[1], e_prev[2],
+            goal_.pose.position.x, goal_.pose.position.y,
+            tf2::getYaw(goal_.pose.orientation));
+        if (e[0] != e_prev[0] || e[1] != e_prev[1] || e[2] != e_prev[2])
+        {
+          ROS_INFO("Goal reverted. Metric: (%.3f, %.3f, %.3f), Grid: (%d, %d, %d)",
+                   goal_raw_.pose.position.x, goal_raw_.pose.position.y, tf2::getYaw(goal_raw_.pose.orientation),
+                   e[0], e[1], e[2]);
+        }
+        goal_ = goal_raw_;
         break;
     }
     const auto ts = boost::chrono::high_resolution_clock::now();
     cost_estim_cache_.create(s, e);
     const auto tnow = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Cost estimation cache generated (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
-
-    if (goal_changed)
-    {
-      clearHysteresis();
-      has_hysteresis_map_ = false;
-    }
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Cost estimation cache generated (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_init_dur",
+        dur,
+        "second"));
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_update_dur",
+        0.0,
+        "second"));
 
     publishDebug();
 
-    goal_updated_ = true;
+    cost_estim_cache_created_ = true;
 
     return true;
   }
@@ -700,12 +731,12 @@ protected:
     const ros::Time now = ros::Time::now();
 
     const int map_update_x_min = static_cast<int>(msg->x);
-    const int map_update_x_max = static_cast<int>(msg->x + msg->width);
+    const int map_update_x_max = std::max(static_cast<int>(msg->x + msg->width) - 1, 0);
     const int map_update_y_min = static_cast<int>(msg->y);
-    const int map_update_y_max = static_cast<int>(msg->y + msg->height);
+    const int map_update_y_max = std::max(static_cast<int>(msg->y + msg->height) - 1, 0);
 
-    if (static_cast<size_t>(map_update_x_max) > map_info_.width ||
-        static_cast<size_t>(map_update_y_max) > map_info_.height ||
+    if (static_cast<size_t>(map_update_x_max) >= map_info_.width ||
+        static_cast<size_t>(map_update_y_max) >= map_info_.height ||
         msg->angle > map_info_.angle)
     {
       ROS_WARN(
@@ -719,21 +750,34 @@ protected:
 
     last_costmap_ = now;
 
-    cm_.copy_partially(cm_base_, Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
-                       Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, static_cast<int>(map_info_.angle)));
-    cm_rough_.copy_partially(cm_rough_base_, Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
-                             Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 1));
-    cm_updates_.clear_partially(-1, Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
-                                Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 1));
+    cm_.copy_partially(
+        cm_base_,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, static_cast<int>(map_info_.angle) - 1));
+    cm_rough_.copy_partially(
+        cm_rough_base_,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 0));
+    cm_updates_.clear_partially(
+        -1,
+        Astar::Vec(prev_map_update_x_min_, prev_map_update_y_min_, 0),
+        Astar::Vec(prev_map_update_x_max_, prev_map_update_y_max_, 0));
 
     // Should search 1px around the region to
     // update the costmap even if the edge of the local map is obstacle
-    const int search_range_x_min = std::max(0, std::min(prev_map_update_x_min_, map_update_x_min) - 1);
-    const int search_range_x_max = std::min(static_cast<int>(map_info_.width),
-                                            std::max(prev_map_update_x_max_, map_update_x_max) + 1);
-    const int search_range_y_min = std::max(0, std::min(prev_map_update_y_min_, map_update_y_min) - 1);
-    const int search_range_y_max = std::min(static_cast<int>(map_info_.height),
-                                            std::max(prev_map_update_y_max_, map_update_y_max) + 1);
+    const int search_range_x_min = std::max(
+        0,
+        std::min(prev_map_update_x_min_, map_update_x_min) - 1);
+    const int search_range_x_max = std::min(
+        static_cast<int>(map_info_.width - 1),
+        std::max(prev_map_update_x_max_, map_update_x_max) + 1);
+    const int search_range_y_min = std::max(
+        0,
+        std::min(prev_map_update_y_min_, map_update_y_min) - 1);
+    const int search_range_y_max = std::min(
+        static_cast<int>(map_info_.height - 1),
+        std::max(prev_map_update_y_max_, map_update_y_max) + 1);
+
     prev_map_update_x_min_ = map_update_x_min;
     prev_map_update_x_max_ = map_update_x_max;
     prev_map_update_y_min_ = map_update_y_min;
@@ -783,7 +827,12 @@ protected:
       }
     }
     const auto ts_cm_init_end = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Costmaps updated (%.4f)", boost::chrono::duration<float>(ts_cm_init_end - ts_cm_init_start).count());
+    const float ts_cm_init_dur = boost::chrono::duration<float>(ts_cm_init_end - ts_cm_init_start).count();
+    ROS_DEBUG("Costmaps updated (%.4f)", ts_cm_init_dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "costmap_dur",
+        ts_cm_init_dur,
+        "second"));
 
     if (clear_hysteresis && has_hysteresis_map_)
     {
@@ -816,7 +865,7 @@ protected:
 
     if (!fast_map_update_)
     {
-      updateGoal(false);
+      createCostEstimCache(false);
       return;
     }
 
@@ -829,7 +878,7 @@ protected:
 
     if (cm_[e] == 100)
     {
-      updateGoal(false);
+      createCostEstimCache(false);
       return;
     }
 
@@ -847,8 +896,16 @@ protected:
     }
     ROS_DEBUG("Cost estimation cache search queue initial size: %lu, capacity: %lu",
               dm_debug.search_queue_size, dm_debug.search_queue_cap);
-    ROS_DEBUG("Cost estimation cache updated (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Cost estimation cache updated (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_update_dur",
+        dur,
+        "second"));
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "distance_map_init_dur",
+        0.0,
+        "second"));
     publishDebug();
   }
   void cbMap(const costmap_cspace_msgs::CSpace3D::ConstPtr& msg)
@@ -875,27 +932,7 @@ protected:
         map_info_.angular_resolution != msg->info.angular_resolution)
     {
       map_info_ = msg->info;
-
-      range_ = static_cast<int>(search_range_ / map_info_.linear_resolution);
-      hist_ignore_range_ = std::lround(hist_ignore_range_f_ / map_info_.linear_resolution);
-      hist_ignore_range_max_ = std::lround(hist_ignore_range_max_f_ / map_info_.linear_resolution);
-      local_range_ = std::lround(local_range_f_ / map_info_.linear_resolution);
-      esc_range_ = std::lround(esc_range_f_ / map_info_.linear_resolution);
-      esc_angle_ = map_info_.angle / 8;
-      tolerance_range_ = std::lround(tolerance_range_f_ / map_info_.linear_resolution);
-      tolerance_angle_ = std::lround(tolerance_angle_f_ / map_info_.angular_resolution);
-      goal_tolerance_lin_ = std::lround(goal_tolerance_lin_f_ / map_info_.linear_resolution);
-      goal_tolerance_ang_ = std::lround(goal_tolerance_ang_f_ / map_info_.angular_resolution);
-      cc_.angle_resolution_aspect_ = 2.0 / tanf(map_info_.angular_resolution);
-
-      model_.reset(
-          new GridAstarModel3D(
-              map_info_,
-              ec_,
-              local_range_,
-              cost_estim_cache_.gridmap(), cm_, cm_hyst_, cm_rough_,
-              cc_, range_));
-
+      resetGridAstarModel();
       ROS_DEBUG("Search model updated");
     }
     else
@@ -959,18 +996,20 @@ protected:
     hyst_updated_cells_.clear();
     has_hysteresis_map_ = false;
 
+    cm_updates_.clear(0);
+
     has_map_ = true;
 
     cm_rough_base_ = cm_rough_;
     cm_base_ = cm_;
     bbf_costmap_.clear();
 
-    prev_map_update_x_min_ = std::numeric_limits<int>::max();
-    prev_map_update_x_max_ = std::numeric_limits<int>::lowest();
-    prev_map_update_y_min_ = std::numeric_limits<int>::max();
-    prev_map_update_y_max_ = std::numeric_limits<int>::lowest();
+    prev_map_update_x_min_ = map_info_.width;
+    prev_map_update_x_max_ = 0;
+    prev_map_update_y_min_ = map_info_.height;
+    prev_map_update_y_max_ = 0;
 
-    updateGoal();
+    createCostEstimCache();
 
     if (map_update_retained_ && map_update_retained_->header.stamp >= msg->header.stamp)
     {
@@ -1037,6 +1076,7 @@ public:
     : nh_()
     , pnh_("~")
     , tfl_(tfbuf_)
+    , parameter_server_(pnh_)
     , cost_estim_cache_(cm_rough_, bbf_costmap_)
     , jump_(tfbuf_)
   {
@@ -1053,6 +1093,7 @@ public:
     pub_start_ = pnh_.advertise<geometry_msgs::PoseStamped>("path_start", 1, true);
     pub_end_ = pnh_.advertise<geometry_msgs::PoseStamped>("path_end", 1, true);
     pub_status_ = pnh_.advertise<planner_cspace_msgs::PlannerStatus>("status", 1, true);
+    pub_metrics_ = pnh_.advertise<neonavigation_metrics_msgs::Metrics>("metrics", 1, false);
     srs_forget_ = neonavigation_common::compat::advertiseService(
         nh_, "forget_planning_cost",
         pnh_, "forget", &Planner3dNode::cbForget, this);
@@ -1186,9 +1227,8 @@ public:
     int num_task;
     pnh_.param("num_search_task", num_task, num_threads * 16);
     as_.setSearchTaskNum(num_task);
-    int num_cost_estim_task;
-    pnh_.param("num_cost_estim_task", num_cost_estim_task, num_threads * 16);
-    cost_estim_cache_.setParams(cc_, num_cost_estim_task);
+    pnh_.param("num_cost_estim_task", num_cost_estim_task_, num_threads * 16);
+    cost_estim_cache_.setParams(cc_, num_cost_estim_task_);
 
     pnh_.param("retain_last_error_status", retain_last_error_status_, true);
     status_.status = planner_cspace_msgs::PlannerStatus::DONE;
@@ -1196,7 +1236,7 @@ public:
     has_map_ = false;
     has_goal_ = false;
     has_start_ = false;
-    goal_updated_ = false;
+    cost_estim_cache_created_ = false;
 
     escaping_ = false;
     cnt_stuck_ = 0;
@@ -1207,6 +1247,102 @@ public:
 
     act_->start();
     act_tolerant_->start();
+
+    // cbParameter() with the inital parameters will be called within setCallback().
+    parameter_server_.setCallback(boost::bind(&Planner3dNode::cbParameter, this, _1, _2));
+  }
+
+  void resetGridAstarModel()
+  {
+    range_ = static_cast<int>(search_range_ / map_info_.linear_resolution);
+    hist_ignore_range_ = std::lround(hist_ignore_range_f_ / map_info_.linear_resolution);
+    hist_ignore_range_max_ = std::lround(hist_ignore_range_max_f_ / map_info_.linear_resolution);
+    local_range_ = std::lround(local_range_f_ / map_info_.linear_resolution);
+    esc_range_ = std::lround(esc_range_f_ / map_info_.linear_resolution);
+    esc_angle_ = map_info_.angle / 8;
+    tolerance_range_ = std::lround(tolerance_range_f_ / map_info_.linear_resolution);
+    tolerance_angle_ = std::lround(tolerance_angle_f_ / map_info_.angular_resolution);
+    goal_tolerance_lin_ = std::lround(goal_tolerance_lin_f_ / map_info_.linear_resolution);
+    goal_tolerance_ang_ = std::lround(goal_tolerance_ang_f_ / map_info_.angular_resolution);
+    cc_.angle_resolution_aspect_ = 2.0 / tanf(map_info_.angular_resolution);
+
+    model_.reset(
+        new GridAstarModel3D(
+            map_info_,
+            ec_,
+            local_range_,
+            cost_estim_cache_.gridmap(), cm_, cm_hyst_, cm_rough_,
+            cc_, range_));
+  }
+
+  void cbParameter(const Planner3DConfig& config, const uint32_t /* level */)
+  {
+    freq_ = config.freq;
+    freq_min_ = config.freq_min;
+    search_timeout_abort_ = config.search_timeout_abort;
+    search_range_ = config.search_range;
+    antialias_start_ = config.antialias_start;
+    costmap_watchdog_ = ros::Duration(config.costmap_watchdog);
+
+    cc_.max_vel_ = config.max_vel;
+    cc_.max_ang_vel_ = config.max_ang_vel;
+    cc_.min_curve_radius_ = config.min_curve_radius;
+    cc_.weight_decel_ = config.weight_decel;
+    cc_.weight_backward_ = config.weight_backward;
+    cc_.weight_ang_vel_ = config.weight_ang_vel;
+    cc_.weight_costmap_ = config.weight_costmap;
+    cc_.weight_costmap_turn_ = config.weight_costmap_turn;
+    cc_.weight_remembered_ = config.weight_remembered;
+    cc_.in_place_turn_ = config.cost_in_place_turn;
+    cc_.hysteresis_max_dist_ = config.hysteresis_max_dist;
+    cc_.hysteresis_expand_ = config.hysteresis_expand;
+    cc_.weight_hysteresis_ = config.weight_hysteresis;
+
+    goal_tolerance_lin_f_ = config.goal_tolerance_lin;
+    goal_tolerance_ang_f_ = config.goal_tolerance_ang;
+    goal_tolerance_ang_finish_ = config.goal_tolerance_ang_finish;
+
+    overwrite_cost_ = config.overwrite_cost;
+    hist_ignore_range_f_ = config.hist_ignore_range;
+    hist_ignore_range_max_f_ = config.hist_ignore_range_max;
+
+    remember_updates_ = config.remember_updates;
+    remember_hit_odds_ = bbf::probabilityToOdds(config.remember_hit_prob);
+    remember_miss_odds_ = bbf::probabilityToOdds(config.remember_miss_prob);
+
+    local_range_f_ = config.local_range;
+    longcut_range_f_ = config.longcut_range;
+    esc_range_f_ = config.esc_range;
+    tolerance_range_f_ = config.tolerance_range;
+    tolerance_angle_f_ = config.tolerance_angle;
+    find_best_ = config.find_best;
+    force_goal_orientation_ = config.force_goal_orientation;
+    temporary_escape_ = config.temporary_escape;
+    fast_map_update_ = config.fast_map_update;
+    max_retry_num_ = config.max_retry_num;
+    sw_wait_ = config.sw_wait;
+
+    cost_estim_cache_.setParams(cc_, num_cost_estim_task_);
+    ec_ = Astar::Vecf(
+        1.0f / cc_.max_vel_,
+        1.0f / cc_.max_vel_,
+        1.0f * cc_.weight_ang_vel_ / cc_.max_ang_vel_);
+
+    if (map_info_.linear_resolution != 0.0 && map_info_.angular_resolution != 0.0)
+    {
+      resetGridAstarModel();
+      const Astar::Vec size2d(static_cast<int>(map_info_.width), static_cast<int>(map_info_.height), 1);
+      const DistanceMap::Params dmp =
+          {
+              .euclid_cost = ec_,
+              .range = range_,
+              .local_range = local_range_,
+              .longcut_range = static_cast<int>(std::lround(longcut_range_f_ / map_info_.linear_resolution)),
+              .size = size2d,
+              .resolution = map_info_.linear_resolution,
+          };
+      cost_estim_cache_.init(model_, dmp);
+    }
   }
 
   GridAstarModel3D::Vec pathPose2Grid(const geometry_msgs::PoseStamped& pose) const
@@ -1224,7 +1360,7 @@ public:
     {
       const ros::Time prev_map_update_stamp = last_costmap_;
       ros::spinOnce();
-      const bool costmap_udpated = last_costmap_ != prev_map_update_stamp;
+      const bool costmap_updated = last_costmap_ != prev_map_update_stamp;
 
       if (has_map_)
       {
@@ -1237,7 +1373,7 @@ public:
           return;
         }
 
-        if (costmap_udpated && previous_path.poses.size() > 1)
+        if (costmap_updated && previous_path.poses.size() > 1)
         {
           for (const auto& path_pose : previous_path.poses)
           {
@@ -1287,14 +1423,19 @@ public:
       const ros::Time now = ros::Time::now();
       next_replan_time = now;
 
-      if (has_map_ && !goal_updated_ && has_goal_)
+      if (has_map_ && !cost_estim_cache_created_ && has_goal_)
       {
-        updateGoal();
+        createCostEstimCache();
       }
       bool has_costmap(false);
       if (costmap_watchdog_ > ros::Duration(0))
       {
-        if (last_costmap_ + costmap_watchdog_ < now)
+        const ros::Duration costmap_delay = now - last_costmap_;
+        metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+            "costmap_delay",
+            costmap_delay.toSec(),
+            "second"));
+        if (costmap_delay > costmap_watchdog_)
         {
           ROS_WARN_THROTTLE(1.0,
                             "Navigation is stopping since the costmap is too old (costmap: %0.3f)",
@@ -1310,6 +1451,10 @@ public:
       }
       else
       {
+        metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+            "costmap_delay",
+            -1.0,
+            "second"));
         has_costmap = true;
       }
 
@@ -1363,6 +1508,7 @@ public:
         }
         else
         {
+          bool skip_path_planning = false;
           if (escaping_)
           {
             status_.error = planner_cspace_msgs::PlannerStatus::PATH_NOT_FOUND;
@@ -1387,32 +1533,52 @@ public:
 
             continue;
           }
+          else if (!cost_estim_cache_created_)
+          {
+            skip_path_planning = true;
+            if (is_start_occupied_)
+            {
+              status_.error = planner_cspace_msgs::PlannerStatus::IN_ROCK;
+            }
+            else
+            {
+              status_.error = planner_cspace_msgs::PlannerStatus::PATH_NOT_FOUND;
+            }
+          }
           else
           {
             status_.error = planner_cspace_msgs::PlannerStatus::GOING_WELL;
           }
-          nav_msgs::Path path;
-          path.header = map_header_;
-          path.header.stamp = now;
-          makePlan(start_.pose, goal_.pose, path, true);
-          if (use_path_with_velocity_)
+
+          if (skip_path_planning)
           {
-            // NaN velocity means that don't care the velocity
-            pub_path_velocity_.publish(
-                trajectory_tracker_msgs::toPathWithVelocity(path, std::numeric_limits<double>::quiet_NaN()));
+            publishEmptyPath();
+            previous_path.poses.clear();
           }
           else
           {
-            pub_path_.publish(path);
-          }
-          previous_path = path;
-
-          if (sw_wait_ > 0.0)
-          {
-            const int sw_index = getSwitchIndex(path);
-            is_path_switchback = (sw_index >= 0);
-            if (is_path_switchback)
-              sw_pos_ = path.poses[sw_index];
+            nav_msgs::Path path;
+            path.header = map_header_;
+            path.header.stamp = now;
+            makePlan(start_.pose, goal_.pose, path, true);
+            if (use_path_with_velocity_)
+            {
+              // NaN velocity means that don't care the velocity
+              pub_path_velocity_.publish(
+                  trajectory_tracker_msgs::toPathWithVelocity(path, std::numeric_limits<double>::quiet_NaN()));
+            }
+            else
+            {
+              pub_path_.publish(path);
+            }
+            previous_path = path;
+            if (sw_wait_ > 0.0)
+            {
+              const int sw_index = getSwitchIndex(path);
+              is_path_switchback = (sw_index >= 0);
+              if (is_path_switchback)
+                sw_pos_ = path.poses[sw_index];
+            }
           }
         }
       }
@@ -1423,9 +1589,26 @@ public:
         publishEmptyPath();
         previous_path.poses.clear();
       }
+      status_.header.stamp = now;
       pub_status_.publish(status_);
       diag_updater_.force_update();
 
+      metrics_.header.stamp = now;
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "stuck_cnt",
+          cnt_stuck_,
+          "count"));
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "error",
+          status_.error,
+          "enum"));
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "status",
+          status_.status,
+          "enum"));
+      pub_metrics_.publish(metrics_);
+      metrics_.data.clear();
+
       if (is_path_switchback)
       {
         next_replan_time += ros::Duration(sw_wait_);
@@ -1458,6 +1641,7 @@ protected:
     if (!cm_.validate(s, range_))
     {
       ROS_ERROR("You are on the edge of the world.");
+      status_.error = planner_cspace_msgs::PlannerStatus::IN_ROCK;
       return false;
     }
 
@@ -1521,7 +1705,7 @@ protected:
         {
           goal_ = goal_raw_;
           escaping_ = false;
-          updateGoal();
+          createCostEstimCache();
           ROS_INFO("Escaped");
           return true;
         }
@@ -1587,7 +1771,7 @@ protected:
           goal_.pose.position.x = x;
           goal_.pose.position.y = y;
 
-          updateGoal();
+          createCostEstimCache();
           return false;
         }
       }
@@ -1644,8 +1828,12 @@ protected:
         return false;
     }
     const auto tnow = boost::chrono::high_resolution_clock::now();
-    ROS_DEBUG("Path found (%0.4f sec.)",
-              boost::chrono::duration<float>(tnow - ts).count());
+    const float dur = boost::chrono::duration<float>(tnow - ts).count();
+    ROS_DEBUG("Path found (%0.4f sec.)", dur);
+    metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+        "path_search_dur",
+        dur,
+        "second"));
 
     geometry_msgs::PoseArray poses;
     poses.header = path.header;
@@ -1722,8 +1910,12 @@ protected:
       }
       has_hysteresis_map_ = true;
       const auto tnow = boost::chrono::high_resolution_clock::now();
-      ROS_DEBUG("Hysteresis map generated (%0.4f sec.)",
-                boost::chrono::duration<float>(tnow - ts).count());
+      const float dur = boost::chrono::duration<float>(tnow - ts).count();
+      ROS_DEBUG("Hysteresis map generated (%0.4f sec.)", dur);
+      metrics_.data.push_back(neonavigation_metrics_msgs::metric(
+          "hyst_map_dur",
+          dur,
+          "second"));
       publishDebug();
     }
 
